package lsedit;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Rectangle;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTree;
import javax.swing.Scrollable;

import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.tree.ExpandVetoException;

import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeWillExpandListener;

/* In a large diagram the m_tree can be huge .. therefore we compute
 * the contents of m_tree only when needed and discard ASAP
 */


/* This class makes TreePaths more efficient by leveraging the fact that the underlying data model is naturally hierarchical */

class MyTreePath extends TreePath
{
	protected static MyTreePath parentPath(EntityInstance e)
	{
		EntityInstance parent = e.getContainedBy();
		if (parent != null) {
			return new MyTreePath(parent);
		}
		return null;
	}

	public MyTreePath(EntityInstance e)
	{
		super(parentPath(e), e);
	}

	public MyTreePath(MyTreePath path, EntityInstance e)
	{
		super(path, e);
	}

    public boolean equals(Object o) 
	{
		if (o instanceof MyTreePath) {
			if (getLastPathComponent() == ((MyTreePath) o).getLastPathComponent()) {
				return true;
		}	}
		return false;
    }

	// Returns true if aTreePath is a descendant of this TreePath. 

	public boolean isDescendant(TreePath aTreePath) 
	{
		EntityInstance e  = (EntityInstance) getLastPathComponent();
		EntityInstance e1 = (EntityInstance) aTreePath.getLastPathComponent();

		return e.hasDescendant(e1);
	}

    public TreePath pathByAddingChild(Object child) 
	{
		return new MyTreePath(this, (EntityInstance) child);
    }
}

public class TextTree extends JComponent implements /* Scrollable, */ TreeModel, MouseListener, MouseMotionListener, TreeWillExpandListener
{
	public    static final String m_helpStr	 = "Right click for menu.";
	
	protected static final String[] m_say    = 
	{
		"Table of contents is hidden",
		"Press Alt-v or click to show TOC"
	};

	protected static final int margin     = 10;
	protected static final int say_height = 30;

	protected static final Color titleColor = Color.red.darker();

	public final static String DEFAULT_TOC_FONT_NAME  = FontCache.DEFAULT_FONT_NAME;
	public final static int    DEFAULT_TOC_FONT_STYLE = Font.PLAIN;
	public final static int    DEFAULT_TOC_FONT_SIZE  = 11;

	protected static Font m_textFont  = FontCache.get(DEFAULT_TOC_FONT_NAME, DEFAULT_TOC_FONT_STYLE, DEFAULT_TOC_FONT_SIZE);

	protected static Hashtable	m_childVectors;

	// Need to subclass since want to pass right click mouse events up chain 

	class MyTree extends JTree implements MouseListener
	{

		// Need subclass since want to have special rendering

		class MyTreeCellRenderer extends DefaultTreeCellRenderer /* extends JLabel */
		{
			public MyTreeCellRenderer()
			{
			}

			public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus)
			{
				Component				c;
				EntityInstance			e;
				int						cnt;

				c      = super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
				if (value instanceof EntityInstance) {
					e    = (EntityInstance) value;
					cnt  = e.numChildren();
					if (cnt > 1) {
						setText(e.getEntityLabel() + " (" + cnt + ")");
					}
				} 
				return(c);
		}	}

		// Always return false so that collapsing marked nodes does not highlight new collapsed node

		protected boolean removeDescendantSelectedPaths(TreePath path, boolean includePath) 
		{
			super.removeDescendantSelectedPaths(path, includePath);
			return(false);
		}

		public MyTree(TreeModel treeModel)
		{
			super(treeModel);

			MyTreeCellRenderer	renderer;

			setBackground(Diagram.boxColor);
			setFont(m_textFont);
			setEditable(false);
			setShowsRootHandles(true);
			renderer = new MyTreeCellRenderer();
			renderer.setBackgroundNonSelectionColor(Diagram.boxColor);
//			renderer.setBackgroundSelectionColor(Color.BLUE); 
			setCellRenderer(renderer);
			setDragEnabled(true);
			addMouseListener(this);
		}

		// MouseListener interface

		public void mouseClicked(MouseEvent ev)
		{
			if (ev.isMetaDown()) {
				int				x, y;
				TreePath		selPath;
				EntityInstance	e;
				Diagram			diagram;
					
				x       = ev.getX();
				y       = ev.getY();
				selPath = getPathForLocation(x, y);
				e       = null;
				if(selPath == null) {
					TextTree.this.mouseClicked(ev);
				} else {
					e = (EntityInstance) selPath.getLastPathComponent();
				}
				doRightPopup(ev, e);
		}	}
	

		public void mouseEntered(MouseEvent ev)
		{
		}

		public void mouseExited(MouseEvent e)
		{
		}

		public void mousePressed(MouseEvent ev)
		{
		}

		public void mouseReleased(MouseEvent ev)
		{
		}
	}

	protected LandscapeEditorCore m_ls;

	protected MyTree			  m_tree;
	protected JPanel			  m_say_panel;
	protected JComponent		  m_use;

	protected JTabbedPane		  m_tabbedPane;
	protected JScrollPane		  m_scrollPane;
	protected JLabel[]			  m_say_labels;

	/* If hidden and first time actually tabbed to this set visible */

	protected boolean			  m_firsttime = true;
	protected Vector			  m_treeModelListeners;

	// --------------
	// Public methods 
	// --------------

	public TextTree(LandscapeEditorCore ls, JTabbedPane tabbedPane) 
	{
		super();

		int						i, cnt;
		JLabel					say;

		setLayout(null);

		m_ls          = ls;
		m_tabbedPane  = tabbedPane;
		m_say_panel   = new JPanel();
//		m_say_panel.setBackground(Color.lightGray);

		m_say_panel.setLayout(new GridLayout(0, 1));

		cnt   = m_say.length;
		m_say_labels = new JLabel[cnt];
		for (i = 0; i < cnt; ++i) {
			m_say_labels[i] = say = new JLabel(m_say[i]);
			say.setBackground(Diagram.boxColor);
			say.setForeground(Color.red);
			say.setFont(m_textFont);
			say.setVisible(true);
			m_say_panel.add(say);
		}
		m_say_panel.setSize(m_say_panel.getPreferredSize());
		fillTOC();

		setToolTipText(m_helpStr);

		m_scrollPane = new JScrollPane();
//		m_scrollPane.setViewport(new MyViewport());
		m_scrollPane.setBounds(0,0,tabbedPane.getWidth(), tabbedPane.getHeight());
		setBounds(0, 0, tabbedPane.getWidth(), tabbedPane.getHeight());
		m_scrollPane.setViewportView(this);

		tabbedPane.addTab("TOC", null, m_scrollPane, m_helpStr);
		addMouseListener(this);
	}
	
	public static Font getTextFont()
	{
		return m_textFont;
	}

	public static void setTextFont(Font font)
	{
		m_textFont = font;
	}

	public void textFontChanged(Font font)
	{
		JLabel	say;
		int		i, cnt;
		
		cnt        = m_say_labels.length;
		for (i = 0; i < cnt; ++i) {
			say = m_say_labels[i];
			say.setFont(font);
		}
		m_tree.setFont(font);
		fillTOC();
	}

	public void paintComponent(Graphics g) 
	{
		int	width                      = getWidth();
		int height                     = getHeight();

		super.paintComponent(g);

		// This is a real cludge -- for some reason
		// The tree does not fill all its area with
		// the desired color when scrolling becomes
		// enabled.  Obscure this fact by painting
		// the entire background here.

		g.setColor(Color.lightGray);
		g.fillRect(0,0,width,height);
	}

	public void firstTime()
	{
		m_firsttime = true;
	}

 	public void activate() 
	{
		m_tabbedPane.setSelectedComponent(m_scrollPane);
	}

	public boolean isActive() 
	{
		Component active;

		if (!isVisible()) {
			return(false);
		}
		active = m_tabbedPane.getSelectedComponent();
		return(active != null && m_scrollPane == active);
	}

	protected JPopupMenu buildPopup(EntityInstance e)
	{
		JPopupMenu	m;
		MyMenuItem	m1;
		Diagram		diagram = m_ls.getDiagram();

		
		m = new JPopupMenu("TOC options");

		if (e != null) {
			m1 = Do.navigateMenuItem(m, m_ls, e);
		}

		if (m_use != null && m_use == m_tree && diagram != null) {
			Clipboard	clipboard = diagram.getClipboard();
			// Get paths of all selected nodes
			TreePath[] paths = m_tree.getSelectionPaths();
			if (paths != null && paths.length > 0) {
				m1 = Do.cutMenuItem(m, m_ls);
				m1.setObject(this);
			} 
			if (clipboard != null && !clipboard.isEmpty()) {
				if (paths != null && paths.length == 1) {
					m1 = Do.pasteMenuItem(m, m_ls);
					m1.setObject(this);
		}	}	}
		Do.openCloseTOCMenuItem(m, m_ls);
		Do.alignTOCMenuItem(m, m_ls);
		m_ls.hideTOCMenuItem(m, m_ls);
		FontCache.setMenuTreeFont(m); 
		return(m);
	}

	// Identify the nodes to cut

	public Vector getGroup()
	{
		Vector	ret = null;

		if (m_use != null && m_use == m_tree) {
			EntityInstance	e;
			int				i, cnt;
			TreePath[]		paths = m_tree.getSelectionPaths();

			if (paths != null && (cnt = paths.length) > 0) {
				ret = new Vector(cnt);
				for (i = 0; i < cnt; ++i) {
					e = (EntityInstance) paths[i].getLastPathComponent();
					ret.add(e);
		}	}	}
		return(ret);
	}

	// Identify the single selected node where we are to paste

	public EntityInstance targetEntity()
	{
		if (m_use != null && m_use == m_tree) {
			EntityInstance	e;
			TreePath[]		paths = m_tree.getSelectionPaths();

			if (paths != null && paths.length == 1) {
				e = (EntityInstance) paths[0].getLastPathComponent();
				return(e);
		}	}	
		return(null);
	}

	// Free all memory used by TOC

	public void emptyTOC()
	{
		removeAll();
		m_use          = null;
		m_tree         = null;
		m_childVectors = null;
	}	

	protected void fillTOC() 
	{
		Diagram			diagram = m_ls.getDiagram();
		EntityInstance	root;
		JComponent		use;

		emptyTOC();

		if (diagram == null) {
			return;
		}
		root = diagram.getRootInstance();
		if (root == null) {
			return;
		}

//		System.out.println("TextTree.fillToc() " + m_ls.isTocHidden());

		if (m_firsttime) {
			if (m_ls.isTocHidden()) {
				m_ls.setTocHiddenState(false);
				m_firsttime = false;
		}	}

		if (m_ls.isTocHidden()) {
			use = m_say_panel;
		} else {
			m_treeModelListeners = null;
			m_tree               = new MyTree((TreeModel) this);
			m_tree.addTreeWillExpandListener(this);

/*
			if (root != null) {
				m_tree.setModel(this);
			}
 */
//			m_tree.setSize(m_tree.getPreferredSize());
			use = m_tree;
		}
		// If we don't do this setSize the tree is not painted
		setSize(use.getPreferredSize());
		add(use);
		m_use = use;
		use.setVisible(true);
		repaint();
	}
		
	public MyTreePath getTreePath(EntityInstance e)
	{
		if (m_use != null && m_use == m_tree) {
			return new MyTreePath(e);
		}
		return null;
	}

	public void insertTOC(EntityInstance e)
	{
		EntityInstance	parent = e.getContainedBy();

		if (parent != null) {

			if (m_childVectors != null) {
				// Make sure we don't remember false information
				m_childVectors.remove(parent.getId());
			}

			MyTreePath path = getTreePath(parent);

			if (path != null) {


				int[]				indices = {	getIndexOfChild(parent, e) };
				Object[]			objects = { e };
				TreeModelEvent		event   = new TreeModelEvent(this, path, indices, objects);
				TreeModelListener	listener;
				int					i;

				for (i = m_treeModelListeners.size(); i > 0; ) {
					listener = (TreeModelListener) m_treeModelListeners.elementAt(--i);
					listener.treeNodesInserted(event);
//					listener.treeStructureChanged(event);
	}	}	}	}

	public void deleteTOC(EntityInstance e)
	{
		EntityInstance	parent = e.getContainedBy();

		if (parent != null) {

			if (m_childVectors != null) {
				// Make sure we don't remember false information
				m_childVectors.remove(parent.getId());
			}

			MyTreePath path = getTreePath(parent);

			if (path != null) {
				int[]				indices = {	getIndexOfChild(parent, e) };
				Object[]			objects = { e };
				TreeModelEvent		event   = new TreeModelEvent(this, path, indices, objects);
				TreeModelListener	listener;
				int					i;

				for (i = m_treeModelListeners.size(); i > 0; ) {
					listener = (TreeModelListener) m_treeModelListeners.elementAt(--i);
					listener.treeNodesRemoved(event);
//					listener.treeStructureChanged(event);
	}	}	}	}

	public void containerCut(EntityInstance pe, EntityInstance e)
	{
		if (pe != null) {

			if (m_childVectors != null) {
				m_childVectors.remove(pe.getId());
				m_childVectors.remove(e.getId());
			}

			MyTreePath path = getTreePath(pe);

			if (path != null) {
				TreeModelEvent		event = new TreeModelEvent(this, path);
				TreeModelListener	listener;
				int					i;

				for (i = m_treeModelListeners.size(); i > 0; ) {
					listener = (TreeModelListener) m_treeModelListeners.elementAt(--i);
					listener.treeStructureChanged(event);
	}	}	}	}

	public void containerUncut(EntityInstance e)
	{
		containerCut(e.getContainedBy(), e);
	}

	public Dimension getPreferredSize()
	{
		/* Silliness -- simply creating JMenuItems can cause logic to try to layout things before we
		 * have setup
		 */

		if (m_use != null) {
			return(m_use.getPreferredSize());
		}
		return new Dimension(0,0);
	}

	public void setBounds(int x, int y, int width, int height)
	{	
		super.setBounds(x, y, width, height);
		if (m_use != null) {
			m_use.setBounds(x, y, width, height);
	}	}
			
	protected void closeAll()
	{
		int			i, cnt;

		cnt = m_tree.getRowCount();
		for (i = cnt; i > 0; ) {
			m_tree.collapseRow(--i);
		}
	}
		
	protected void expandTo(EntityInstance e)
	{
		int				cnt;
		TreePath		path;
		EntityInstance	e1;

		closeAll();
		m_tree.clearSelection();
		if (e != null) {
			for (cnt = 0; cnt < m_tree.getRowCount(); ++cnt) {
				path     = m_tree.getPathForRow(cnt);
				e1       = (EntityInstance) path.getLastPathComponent();
				if (e1 != null) {
					if (e1 == e) {
						m_tree.setSelectionRow(cnt);
						m_tree.scrollRowToVisible(cnt);
						// Also open this
						m_tree.expandRow(cnt);
						break;
					}
					if (e1.hasDescendant(e)) {
						m_tree.expandRow(cnt);
	}	}	}	}	} 
		
	public void toc_path()
	{
		Diagram diagram = m_ls.getDiagram();
		EntityInstance e;

		if (diagram != null) {
			m_ls.setTocHidden(false);
			activate();
			e = diagram.getDrawRoot();
			expandTo(e);
	}	}

	protected void expandAll()
	{
		int	cnt;

		for (cnt = 0; cnt < m_tree.getRowCount(); ++cnt) {
			m_tree.expandRow(cnt);
	}	}

	// Open up or close down the Table of contents

	public void switch_TOC()
	{
		int	cnt;

		m_ls.setTocHidden(false);
		activate();
		cnt = m_tree.getRowCount();
		if (cnt > 1) {
			closeAll();
		} else {
			expandAll();
		}
	}	

	void rebuildTOC()
	{
		if (m_use != null && m_use == m_tree) {
			fillTOC();
		}
	}

	protected void doRightPopup(MouseEvent ev, EntityInstance e)
	{
		int				x, y;
		JPopupMenu		popupMenu;
				
		x         = ev.getX();
		y         = ev.getY();
		popupMenu = buildPopup(e);
		add(popupMenu);
		popupMenu.show(this, x, y);
		remove(popupMenu);
	}

	// TreeModel interface

	// Adds a listener for the TreeModelEvent posted after the tree changes. 

	public void addTreeModelListener(TreeModelListener l) 
	{
		if (m_treeModelListeners == null) {
			m_treeModelListeners = new Vector();
		}
		m_treeModelListeners.add(l);
	}

	// Removes a listener previously added with addTreeModelListener. 

	public void removeTreeModelListener(TreeModelListener l) 
	{
		if (m_treeModelListeners != null) {
			m_treeModelListeners.remove(l);
			if (m_treeModelListeners.isEmpty()) {
				m_treeModelListeners = null;
	}	}	}

	// Returns the number of children of parent. 

	public int getChildCount(Object parent) 
	{
		return ((EntityInstance) parent).numChildren();
	}

	protected Vector sortedChildren(EntityInstance e)
	{
		if (m_childVectors == null) {
			m_childVectors = new Hashtable(1024);
		} 

		String			id       = e.getId();
		Vector			children = (Vector) m_childVectors.get(id);

		if (children == null) {
			children = new Vector();
			e.addChildren(children);
			SortVector.byString(children, true);
			m_childVectors.put(id, children);
		}
		return children;
	}

	// Returns the child of parent at index index in the parent's child array. 

	public Object getChild(Object parent, int index)
	{
		EntityInstance	e        = (EntityInstance) parent;

		if (!m_ls.isSortTOC()) {
			return e.getChild(index);
		}

		Vector			children = sortedChildren(e);
		
		return children.elementAt(index);
	}

    // Returns the index of child in parent. 

	public int getIndexOfChild(Object parent, Object child) 
	{
		if (parent != null) {

			EntityInstance	e        = (EntityInstance) parent;

			if (!m_ls.isSortTOC()) {
				return e.getIndexOfChild(child);
			}

			Vector			children = sortedChildren(e);
			int				index;
		
			for (index = children.size(); index > 0; ) {
				e = (EntityInstance) children.elementAt(--index);
				if (e == child) {
					return index;
		}	}	}
		return -1;
	}

	// Returns the root of the tree. 

	public Object getRoot() 
	{
		Diagram			diagram = m_ls.getDiagram();
		EntityInstance	root;

		if (diagram == null) {
			return null;
		}
		return diagram.getRootInstance();
	}

	// Returns true if node is a leaf. 

	public boolean isLeaf(Object node) 
	{
		return (getChildCount(node) == 0);
	}

	// Messaged when the user has altered the value for the item identified by path to newValue. 

	public void valueForPathChanged(TreePath path, Object newValue)
	{
	} 

	// Generic 	MouseListener interface

	public void mouseClicked(MouseEvent ev)
	{
		
		if (!ev.isMetaDown()) {
			if (m_ls.isTocHidden()) {
				m_ls.setTocHidden(false);
			}
		} else {
			doRightPopup(ev, null);
	}	}

	public void mouseEntered(MouseEvent ev)
	{
	}

	public void mouseExited(MouseEvent e)
	{
	}

	public void mousePressed(MouseEvent ev)
	{
	}

	public void mouseReleased(MouseEvent ev)
	{
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
	}

	public void mouseMoved(MouseEvent ev)
	{
	}

	// TreeWillExpandListener implementation 

	public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException
    {
    }
                                                                                                                                                                                    
    //Required by TreeWillExpandListener interface.
    
	public void treeWillCollapse(TreeExpansionEvent event)
    {
		if (m_childVectors != null) {
			EntityInstance  e  = (EntityInstance)  event.getPath().getLastPathComponent();
			String			id = e.getId();

			m_childVectors.remove(id);
    }	}

/*
	// Scrollable interface
	// We need to implement this to stop iterative size changes causing repositioning to (0,0)

	public Dimension getPreferredScrollableViewportSize()
	{
		return(getSize());
	}
	 
	public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction)
	{
		return(50);
	}
	 
	public boolean getScrollableTracksViewportHeight() 
	{
		return(false);
	}

	public boolean getScrollableTracksViewportWidth() 
	{
		return(false);
	}

	public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction)
	{
		return(10);
	} 
*/
	public String toString()
	{
		return("TextTree");
	}
}





